import pytest

from yadm import fields
from yadm.documents import Document


@pytest.mark.parametrize('input, output', [
    ('3.14', '3.14'),
    ('0.14', '0.14'),
    ('30.14', '30.14'),
    ('3.1', '3.10'),
    ('0.1', '0.10'),
])
def test_money_str(input, output):
    assert str(fields.Money(input)) == output


@pytest.mark.parametrize('input, output', [
    ('3.14', 314),
    ('0.14', 14),
    ('30.14', 3014),
    ('3.1', 310),
    ('0.1', 10),
])
@pytest.mark.parametrize('func', [
    lambda money: money.total_cents,
    lambda money: money.to_mongo(),  # b/c function, for coverage
])
def test_money_cents(func, input, output):
    assert func(fields.Money(input)) == output


class Doc(Document):
    __collection__ = 'testdocs'
    money = fields.MoneyField()


def test_save(db):  # noqa
    doc = Doc()
    doc.money = fields.Money('3.14')
    db.save(doc)

    data = db.db.testdocs.find_one()

    assert 'money' in data
    assert isinstance(data['money'], int)
    assert data['money'] == 314


def test_load(db):
    db.db.testdocs.insert({'money': 314})

    doc = db.get_queryset(Doc).find_one()

    assert hasattr(doc, 'money')
    assert isinstance(doc.money, fields.Money)
    assert doc.money == fields.Money('3.14')


def test_load_and_save(db):
    db.db.testdocs.insert({'money': 314})

    doc = db.get_queryset(Doc).find_one()
    doc.money = fields.Money('42.00')
    db.save(doc)

    assert hasattr(doc, 'money')
    assert isinstance(doc.money, fields.Money)
    assert doc.money == fields.Money('42.00')

    data = db.db.testdocs.find_one()

    assert 'money' in data
    assert isinstance(data['money'], int)
    assert data['money'] == 4200



from datetime import datetime

import pytz

from yadm import fields
from yadm.documents import Document


class Doc(Document):
    __collection__ = 'testdocs'
    dt = fields.DatetimeField()
    now = fields.DatetimeField(auto_now=True)


def test_cast():
    doc = Doc()
    doc.dt = datetime(1970, 1, 1, tzinfo=pytz.utc).isoformat()
    assert isinstance(doc.dt, datetime)


def test_default_none():
    doc = Doc()
    assert not hasattr(doc, 'dt')


def test_save(db):
    doc = Doc()
    doc.dt = datetime(1970, 1, 1, tzinfo=pytz.utc)
    db.insert(doc)

    doc = db.get_queryset(Doc).find_one(doc.id)
    data = db.db.testdocs.find_one()

    assert isinstance(data['dt'], datetime)
    assert data['dt'] == doc.dt


def test_load(db):
    epoch = datetime(1970, 1, 1, tzinfo=pytz.utc)
    _id = db.db.testdocs.insert({'dt': epoch})

    doc = db.get_queryset(Doc).find_one(_id)

    assert isinstance(doc.dt, datetime)
    assert doc.dt == epoch


def test_default_auto_now():
    doc = Doc()
    assert hasattr(doc, 'now')
    assert isinstance(doc.now, datetime)


def test_default_now_save(db):
    doc = Doc()
    db.insert(doc)

    doc = db.get_queryset(Doc).find_one(doc.id)
    data = db.db.testdocs.find_one()

    assert isinstance(data['now'], datetime)
    assert data['now'] == doc.now


def test_now_save(db):
    epoch = datetime(1970, 1, 1, tzinfo=pytz.utc)

    doc = Doc()
    doc.now = epoch
    db.insert(doc)

    doc = db.get_queryset(Doc).find_one(doc.id)
    data = db.db.testdocs.find_one()

    assert isinstance(data['now'], datetime)
    assert data['now'] == epoch


def test_now_load(db):
    epoch = datetime(1970, 1, 1, tzinfo=pytz.utc)
    _id = db.db.testdocs.insert({'now': epoch})

    doc = db.get_queryset(Doc).find_one(_id)

    assert isinstance(doc.now, datetime)
    assert doc.now == epoch



"""
Work with embedded documents.

.. code-block:: python

    class EDoc(EmbeddedDocument):
        i = fields.IntegerField()

    class Doc(Document):
        __collection__ = 'docs'
        edoc = EmbeddedDocumentField(EDoc)

    doc = Doc()
    doc.edoc = EDoc()
    doc.edoc.i = 13
    db.insert(doc)
"""

from yadm.common import EnclosedDocDescriptor
from yadm.markers import AttributeNotSet
from yadm.fields.base import Field, pass_null
from yadm.serialize import to_mongo, from_mongo
from yadm.testing import create_fake


class EmbeddedDocumentField(Field):
    """ Field for embedded objects.

    :param EmbeddedDocument embedded_document_class:
        class for embedded document
    :param bool auto_create: automatic creation embedded
        document from access
    """
    auto_create = True
    embedded_document_class = EnclosedDocDescriptor('embedded')

    def __init__(self, embedded_document_class, *,
                 auto_create=True, **kwargs):
        super().__init__(**kwargs)
        self.embedded_document_class = embedded_document_class
        self.auto_create = auto_create

    def get_if_attribute_not_set(self, document):
        """ Call if key not exist in document.

        If auto_create is True, create and return new
        embedded document. Else AttributeError is raised.
        """
        if self.auto_create:
            return self.embedded_document_class(
                __parent__=document,
                __name__=self.name,
            )
        else:
            raise AttributeError(self.name)

    def get_fake(self, document, faker, depth):
        return create_fake(
            self.embedded_document_class,
            __parent__=document,
            __name__=self.name,
            __faker__=faker,
            __depth__=depth)

    @pass_null
    def prepare_value(self, document, value):
        if value is AttributeNotSet:
            return value

        elif isinstance(value, dict):
            value = self.embedded_document_class(
                __parent__=document,
                __name__=self.name,
                **value)

        elif isinstance(value, self.embedded_document_class):
            value.__parent__ = document
            value.__name__ = self.name

        elif not isinstance(value, self.embedded_document_class):
            raise TypeError('Only {!r}, dict or None is alowed, but {!r} given'
                            ''.format(self.embedded_document_class, type(value)))

        return value

    @pass_null
    def to_mongo(self, document, value):
        return to_mongo(value)

    @pass_null
    def from_mongo(self, document, value):
        value = from_mongo(self.embedded_document_class, value,
                           parent=document, name=self.name)

        return value

    def copy(self):
        """ Return copy of field.
        """
        return self.__class__(self.embedded_document_class,
                              smart_null=self.smart_null)
